from google.cloud.bigquery import Client, LoadJobConfig, WriteDisposition
from google.oauth2 import service_account
from mage_ai.io.base import BaseSQLDatabase, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from pandas import DataFrame
from typing import Mapping


class BigQuery(BaseSQLDatabase):
    """
    Handles data transfer betwee a BigQuery data warehouse and the Mage app.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initializes settings for connecting to a BigQuery warehouse.

        To authenticate (and authorize) access to a BigQuery warehouse, credentials must be provided.
        Below are the different ways in which the BigQuery data loader can access those
        credentials:
        - Define the `GOOGLE_APPLICATION_CREDENTIALS` environment variable to point to a
          service account key. In this case no other no other parameters need to be
          specified.
        - Manually pass in the `google.oauth2.service_account.Credentials` object with the
        keyword argument `credentials`
        - Manually pass in the path to the credentials with the keyword argument
        `path_to_credentials`.
        - Manually define the service key key-value set to use (such as a dictionary containing
        the same parameters as a service key) with the keyword argument `credentials_mapping`

        All keyword arguments except for `path_to_credentials` and `credentials_mapping` will be passed
        to the Google BigQuery client, accepting all other configuration settings there.
        """
        verbose = kwargs.get('verbose')
        if verbose is not None:
            kwargs.pop('verbose')
        super().__init__(verbose=verbose)

        credentials = kwargs.get('credentials')
        if credentials is None:
            if 'credentials_mapping' in kwargs:
                mapping_obj = kwargs.pop('credentials_mapping')
                if mapping_obj is not None:
                    credentials = service_account.Credentials.from_service_account_info(mapping_obj)
            if 'path_to_credentials' in kwargs:
                path = kwargs.pop('path_to_credentials')
                if path is not None:
                    credentials = service_account.Credentials.from_service_account_file(path)
            if 'credentials' in kwargs:
                kwargs.pop('credentials')
        with self.printer.print_msg(f'Connecting to BigQuery warehouse'):
            self.client = Client(credentials=credentials, **kwargs)

    def load(self, query_string: str, limit: int = QUERY_ROW_LIMIT, **kwargs) -> DataFrame:
        """
        Loads data from BigQuery into a Pandas data frame based on the query given.
        This will fail if the query returns no data from the database. When a select query
        is provided, this function will load at maximum 100,000 rows of data. To operate on more data,
        consider performing data transformations in warehouse.



        Args:
            query_string (str): Query to fetch a table or subset of a table.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to 100000.
            **kwargs: Additional arguments to pass to query, such as query configurations

        Returns:
            DataFrame: Data frame associated with the given query.
        """
        with self.printer.print_msg(f'Loading data frame with query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            return self.client.query(
                self._enforce_limit(query_string, limit), *kwargs
            ).to_dataframe()

    def export(
        self,
        df: DataFrame,
        table_id: str,
        if_exists: str = 'replace',
        **configuration_params,
    ) -> None:
        """
        Exports a data frame to a Google BigQuery warehouse.  If table doesn't
        exist, the table is automatically created.

        Args:
            df (DataFrame): Data frame to export
            table_id (str): ID of the table to export the data frame to. If of the format
            `"your-project.your_dataset.your_table_name"`. If this table exists,
            the table schema must match the data frame schema. If this table doesn't exist,
            the table schema is automatically inferred.
            if_exists (str): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table. In this case the schema must match the original table.
            Defaults to `'replace'`. If `write_disposition` is specified as a keyword argument, this parameter
            is ignored (as both define the same functionality).
            **configuration_params: Configuration parameters for export job
        """
        with self.printer.print_msg(f'Exporting data frame to table \'{table_id}\''):
            config = LoadJobConfig(**configuration_params)
            if 'write_disposition' not in configuration_params:
                if if_exists == 'replace':
                    config.write_disposition = WriteDisposition.WRITE_TRUNCATE
                elif if_exists == 'append':
                    config.write_disposition = WriteDisposition.WRITE_APPEND
                elif if_exists == 'fail':
                    config.write_disposition = WriteDisposition.WRITE_EMPTY
                else:
                    raise ValueError(
                        f'Invalid policy specified for handling existence of table: \'{if_exists}\''
                    )
            self.client.load_table_from_dataframe(df, table_id, job_config=config).result()

    def execute(self, query_string: str, **kwargs) -> None:
        """
        Sends query to the connected BigQuery warehouse.

        Args:
            query_string (str): Query to execute on the BigQuery warehouse.
            **kwargs: Additional arguments to pass to query, such as query configurations
        """
        with self.printer.print_msg(f'Executing query \'{query_string}\''):
            self.client.query(query_string, **kwargs)

    @classmethod
    def with_config(cls, config: BaseConfigLoader, **kwargs) -> 'BigQuery':
        """
        Initializes BigQuery client from configuration loader

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        if ConfigKey.GOOGLE_SERVICE_ACC_KEY in config:
            kwargs['credentials_mapping'] = config[ConfigKey.GOOGLE_SERVICE_ACC_KEY]
        elif ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH in config:
            kwargs['path_to_credentials'] = config[ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH]
        else:
            ValueError(
                'No valid configuration settings found for Google BigQuery. You must specify '
                'either your service account key or the filepath to your service account key.'
            )
        return cls(**kwargs)

    @classmethod
    def with_credentials_file(cls, path_to_credentials: str, **kwargs) -> 'BigQuery':
        """
        Constructs BigQuery data loader using the file containing the service account key.

        Args:
            path_to_credentials (str): Path to the credentials file.

        Returns:
            BigQuery: BigQuery data loader
        """
        return cls(path_to_credentials=path_to_credentials, **kwargs)

    @classmethod
    def with_credentials_object(cls, credentials: Mapping[str, str], **kwargs) -> 'BigQuery':
        """
        Constructs BigQuery data loader using manually specified authentication credentials object.

        Args:
            credentials (Mapping[str, str]): Credentials object. Must contain all the OAuth information necessary
            to authenticate and authorize BigQuery access. This should resemble a service account key.

        Returns:
            BigQuery: BigQuery data loader
        """
        return cls(credentials_mapping=credentials, **kwargs)



from distutils.dir_util import copy_tree
from mage_ai.data_preparation.models.constants import (
    BlockType,
    CUSTOM_EXECUTION_BLOCK_TYPES,
    PIPELINES_FOLDER,
)
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.repo_manager import get_repo_config, get_repo_path
from mage_ai.data_preparation.variable_manager import get_global_variables
from mage_ai.server.active_kernel import (
    get_active_kernel_client,
    get_active_kernel_name,
    switch_active_kernel,
)
from mage_ai.server.execution_manager import (
    cancel_pipeline_execution,
    delete_pipeline_copy_config,
    reset_execution_manager,
    set_current_pipeline_process,
    set_previous_config_path,
)
from mage_ai.server.kernel_output_parser import DataType
from mage_ai.server.kernels import KernelName
from mage_ai.server.utils.output_display import (
    add_internal_output_info,
    add_execution_code,
    get_block_output_process_code,
    get_pipeline_execution_code,
)
from mage_ai.shared.hash import merge_dict
from jupyter_client import KernelClient
from typing import Callable, Dict
import json
import multiprocessing
import os
import tornado.websocket
import traceback
import uuid


class WebSocketServer(tornado.websocket.WebSocketHandler):
    """Simple WebSocket handler to serve clients."""

    # Note that `clients` is a class variable and `send_message` is a
    # classmethod.
    clients = set()
    running_executions_mapping = dict()

    def open(self):
        WebSocketServer.clients.add(self)

    def on_close(self):
        WebSocketServer.clients.remove(self)

    def check_origin(self, origin):
        return True

    def init_kernel_client(self, kernel_name) -> KernelClient:
        if kernel_name != get_active_kernel_name():
            switch_active_kernel(kernel_name)

        return get_active_kernel_client()

    def on_message(self, raw_message):
        message = json.loads(raw_message)
        output = message.get('output')
        if output:
            self.send_message(output)
            return
        global_vars = message.get('global_vars')
        cancel_pipeline = message.get('cancel_pipeline')
        execute_pipeline = message.get('execute_pipeline')
        kernel_name = message.get('kernel_name', get_active_kernel_name())
        block_uuid = message.get('uuid')
        pipeline_uuid = message.get('pipeline_uuid')
        pipeline = Pipeline(pipeline_uuid, get_repo_path())

        global_vars = message.get('global_vars', get_global_variables(pipeline_uuid))

        value = dict(
            block_uuid=block_uuid,
            pipeline_uuid=pipeline_uuid,
        )

        def publish_message(
            message: str,
            execution_state: str = 'busy',
            msg_type: str = 'stream_pipeline',
            block_uuid: str = None,
        ) -> None:
            msg_id = str(uuid.uuid4())
            value['block_uuid'] = block_uuid
            WebSocketServer.running_executions_mapping[msg_id] = value
            self.send_message(
                dict(
                    data=message,
                    execution_state=execution_state,
                    msg_id=msg_id,
                    msg_type=msg_type,
                    type=DataType.TEXT_PLAIN,
                )
            )

        if cancel_pipeline:
            cancel_pipeline_execution(pipeline, publish_message)
        elif execute_pipeline:
            self.__execute_pipeline(
                pipeline,
                kernel_name,
                global_vars,
                publish_message,
            )
        else:
            self.__execute_block(
                message,
                pipeline,
                kernel_name,
                global_vars,
            )

    @classmethod
    def send_message(self, message: dict) -> None:
        msg_id = message.get('msg_id')
        if msg_id is None:
            return
        msg_id_value = WebSocketServer.running_executions_mapping.get(msg_id, dict())
        block_uuid = msg_id_value.get('block_uuid')
        pipeline_uuid = msg_id_value.get('pipeline_uuid')

        output_dict = dict(uuid=block_uuid, pipeline_uuid=pipeline_uuid)

        message_final = merge_dict(
            message,
            output_dict,
        )

        print(
            f'[{block_uuid}] Sending message for {msg_id} to '
            f'{len(self.clients)} client(s): {message_final}'
        )

        for client in self.clients:
            client.write_message(json.dumps(message_final))


    def __execute_block(
        self,
        message: Dict[str, any],
        pipeline: Pipeline,
        kernel_name: str,
        global_vars: Dict[str, any],
    ) -> None:
        block_type = message.get('type')
        block_uuid = message.get('uuid')
        custom_code = message.get('code')
        run_downstream = message.get('run_downstream')
        run_tests = message.get('run_tests')
        run_upstream = message.get('run_upstream')

        pipeline_uuid = pipeline.uuid

        widget = BlockType.CHART == block_type

        block = pipeline.get_block(block_uuid, widget=widget)
        code = custom_code

        client = self.init_kernel_client(kernel_name)

        value = dict(block_uuid=block_uuid)

        if not custom_code and BlockType.SCRATCHPAD == block_type:
            msg_id = client.execute('')

            WebSocketServer.running_executions_mapping[msg_id] = value

            self.send_message(
                dict(
                    data='',
                    execution_state='idle',
                    msg_id=msg_id,
                    type=DataType.TEXT_PLAIN,
                ),
            )
        else:
            if block is not None and block.type in CUSTOM_EXECUTION_BLOCK_TYPES:
                if kernel_name == KernelName.PYSPARK and not widget:
                    remote_execution = True
                else:
                    remote_execution = False
                code = add_execution_code(
                    pipeline_uuid,
                    block_uuid,
                    custom_code,
                    global_vars,
                    analyze_outputs=False if remote_execution else True,
                    block_type=block_type,
                    kernel_name=kernel_name,
                    pipeline_config=pipeline.get_config_from_yaml(),
                    repo_config=get_repo_config().to_dict(remote=remote_execution),
                    run_tests=run_tests,
                    run_upstream=run_upstream,
                    update_status=False if remote_execution else True,
                    widget=widget,
                )
            
            msg_id = client.execute(add_internal_output_info(code))

            WebSocketServer.running_executions_mapping[msg_id] = value

            block_output_process_code = get_block_output_process_code(
                pipeline_uuid,
                block_uuid,
                block_type=block_type,
                kernel_name=kernel_name,
            )
            if block_output_process_code is not None:
                client.execute(block_output_process_code)

            if run_downstream:
                for block in block.downstream_blocks:
                    self.on_message(json.dumps(dict(
                        code=block.file.content(),
                        pipeline_uuid=pipeline_uuid,
                        type=block.type,
                        uuid=block.uuid,
                    )))


    def __execute_pipeline(
        self,
        pipeline: Pipeline,
        kernel_name: str,
        global_vars: Dict[str, any],
        publish_message: Callable,
    ) -> None:
        pipeline_uuid = pipeline.uuid

        if kernel_name == KernelName.PYSPARK:
            code = get_pipeline_execution_code(
                pipeline_uuid,
                global_vars=global_vars,
                kernel_name=kernel_name,
                pipeline_config=pipeline.to_dict(include_content=True),
                repo_config=get_repo_config().to_dict(remote=True),
                update_status=False if kernel_name == KernelName.PYSPARK else True,
            )
            client = self.init_kernel_client(kernel_name)
            msg_id = client.execute(code)

            WebSocketServer.running_executions_mapping[msg_id] = dict(
                pipeline_uuid=pipeline_uuid
            )
        else:
            # TODO: save config for other kernel types.
            def save_pipeline_config() -> str:
                pipeline_copy = f'{pipeline.uuid}_{str(uuid.uuid4())}'
                new_pipeline_directory = os.path.join(pipeline.repo_path, PIPELINES_FOLDER, pipeline_copy)
                copy_tree(pipeline.dir_path, new_pipeline_directory)
                set_previous_config_path(new_pipeline_directory)
                return new_pipeline_directory

            config_copy_path = save_pipeline_config()
            
            def run_pipeline() -> None:
                try:
                    pipeline.execute_sync(
                        log_func=publish_message,
                    )
                    publish_message(
                        f'Pipeline {pipeline.uuid} execution complete.\n'
                        'You can see the code block output in the corresponding code block.',
                        execution_state = 'idle',
                    )
                except Exception:
                    trace = traceback.format_exc().splitlines()
                    publish_message(f'Pipeline {pipeline.uuid} execution failed with error:')
                    publish_message(trace, 'idle')

                delete_pipeline_copy_config(config_copy_path)

            reset_execution_manager()

            # The pipeline state can potentially break when the execution is cancelled,
            # so we save the pipeline config before execution if the user cancels the excecution.
            save_pipeline_config()
            proc = multiprocessing.Process(target=run_pipeline)
            proc.start()
            set_current_pipeline_process(proc)



import asyncio
import os


def main():
    import sys

    try:
        command = sys.argv[1]
    except IndexError:
        command = 'help'
    if command == 'help':
        print("""Usage:
    mage <command> 

Commands:
    init <project_path>                 Initialize Mage project.
    args:
        project_path                    path of the Mage project to be created.

    start <project_path> [host] [port]  Start Mage server and UI.
    args:
        project_path                    path of the Mage project to be loaded.
        host                            optional argument to specify the host, defaults to localhost
        port                            optional argument to specify the port, defaults to 6789

    run <project_path> <pipeline_uuid>  Run pipeline.
    args:
        project_path                    path of the Mage project that contains the pipeline.
        pipeline_uuid                   uuid of the pipeline to be run.

    test <project_path> <pipeline_uuid> Run pipeline and output tests.
    args:
        project_path                    path of the Mage project that contains the pipeline.
        pipeline_uuid                   uuid of the pipeline to be run and tested.

    create_spark_cluster <project_path>
      args:
        project_path                    path of the Mage project that contains the EMR config.
        """)
    elif command == 'init':
        from mage_ai.data_preparation.repo_manager import init_repo

        repo_path = os.path.join(os.getcwd(), sys.argv[2])
        init_repo(repo_path)
    elif command == 'start':
        from mage_ai.server.server import main as start_server

        host = None
        port = None
        if len(sys.argv) >= 3:
            repo_path = os.path.join(os.getcwd(), sys.argv[2])
            if len(sys.argv) >= 4:
                host = sys.argv[3]
            if len(sys.argv) >= 5:
                port = sys.argv[4]
        else:
            repo_path = os.getcwd()

        asyncio.run(start_server(
            host=host,
            port=port,
            project=repo_path,
        ))
    elif command == 'run':
        from mage_ai.data_preparation.models.pipeline import Pipeline
        from mage_ai.data_preparation.pipeline_executor import PipelineExecutor

        project_path = sys.argv[2]
        pipeline_uuid = sys.argv[3]
        project_path = os.path.abspath(project_path)
        sys.path.append(os.path.dirname(project_path))
        pipeline = Pipeline(pipeline_uuid, project_path)

        PipelineExecutor.get_executor(pipeline).execute(
            analyze_outputs=False,
            update_status=False,
        )
    elif command == 'test':
        from mage_ai.data_preparation.models.pipeline import Pipeline

        project_path = sys.argv[2]
        pipeline_uuid = sys.argv[3]
        project_path = os.path.abspath(project_path)
        sys.path.append(os.path.dirname(project_path))
        pipeline = Pipeline(pipeline_uuid, project_path)

        asyncio.run(
            pipeline.execute(
                analyze_outputs=False,
                run_tests=True,
                update_status=False,
            )
        )
    elif command == 'create_spark_cluster':
        from mage_ai.services.emr.launcher import create_cluster

        project_path = os.path.abspath(sys.argv[2])
        create_cluster(project_path)
    else:
        print(f'Unknown command "{command}". Type "mage help" to see what commands are available.')


if __name__ == "__main__":
    main()



from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.data_preparation.executors.ecs_block_executor import EcsBlockExecutor
from mage_ai.data_preparation.executors.k8s_block_executor import K8sBlockExecutor
from mage_ai.data_preparation.executors.pipeline_executor import PipelineExecutor
from mage_ai.data_preparation.executors.pyspark_block_executor import PySparkBlockExecutor
from mage_ai.data_preparation.executors.pyspark_pipeline_executor import PySparkPipelineExecutor
from mage_ai.data_preparation.models.constants import ExecutorType, PipelineType
from mage_ai.data_preparation.models.pipeline import Pipeline


class ExecutorFactory:
    @classmethod
    def get_pipeline_executor(
        self,
        pipeline: Pipeline,
        executor_type: ExecutorType = None
    ) -> PipelineExecutor:
        if executor_type is None:
            if pipeline.type == PipelineType.PYSPARK:
                executor_type = ExecutorType.PYSPARK
            else:
                executor_type = ExecutorType.LOCAL_PYTHON
        if executor_type == ExecutorType.PYSPARK:
            return PySparkPipelineExecutor(pipeline)
        else:
            return PipelineExecutor(pipeline)

    @classmethod
    def get_block_executor(
        self,
        pipeline: Pipeline,
        block_uuid: str,
        executor_type: ExecutorType = None
    ) -> BlockExecutor:
        if executor_type is None:
            if pipeline.type == PipelineType.PYSPARK:
                executor_type = ExecutorType.PYSPARK
            else:
                executor_type = ExecutorType.LOCAL_PYTHON
        if executor_type == ExecutorType.PYSPARK:
            return PySparkBlockExecutor(pipeline, block_uuid)
        elif executor_type == ExecutorType.ECS:
            return EcsBlockExecutor(pipeline, block_uuid)
        elif executor_type == ExecutorType.K8S:
            return K8sBlockExecutor(pipeline, block_uuid)
        else:
            return BlockExecutor(pipeline, block_uuid)



from typing import Dict


class BlockExecutor:
    def __init__(self, pipeline, block_uuid):
        self.pipeline = pipeline
        self.block_uuid = block_uuid
        self.block = self.pipeline.get_block(block_uuid)

    def execute(
        self,
        analyze_outputs: bool = False,
        global_vars: Dict = None,
        update_status: bool = False,
    ) -> None:
        self.block.execute_sync(
            analyze_outputs=analyze_outputs,
            global_vars=global_vars,
            update_status=update_status,
        )
